import casadi.*

% Time horizon
T = 10;

% State
x1=SX.sym('x1');
x2=SX.sym('x2');
x = [x1; x2];

% Control
u = SX.sym('u');

% ODE right hand side
xdot = [(1 - x2^2)*x1 - x2 + u; x1];

% Quadratic objective
L = x1^2 + x2^2 + u^2;

% Costate
lam = SX.sym('lam',2);

% Hamiltonian function
H = dot(lam,xdot) + L;
display(H)

% Costate equations
ldot = -gradient(H,x);

% H is of a convex quadratic form in u: H = u*u + p*u + q, let's get the coefficient p
p = gradient(H,u);    % this gives us 2*u + p
p = substitute(p,u,0); % replace u with zero: gives us p

% H's unconstrained minimizer is: u = -p/2
u_opt = -p/2;

% We must constrain u to the interval [-1.0, 1.0], convexity of H ensures that the optimum is obtain at the bound when u_opt is outside the interval
u_opt = min(u_opt,  1.0);
u_opt = max(u_opt, -1.0);
display(u_opt)

% Augment the state and substitute in the value for the optimal control
aug = [x;lam];
augdot = [xdot;ldot];
augdot = substitute(augdot,u,u_opt);

% Create an integrator (CVodes)
tf = SX.sym('tf');
dae = struct('x', aug, 'p', tf, 'ode', tf*augdot);
opts = struct('abstol', 1e-8, 'reltol', 1e-8);
F = integrator('F', 'cvodes', dae, opts);

% The initial state
x_init = [0; 1];

% The initial costate
l_init = MX.sym('l_init',2);

% Get an expression for the augmented state at the end time
X = [x_init; l_init];
Fj = F(struct('x0', X, 'p', T));
X = Fj.xf;

% Costate at the final time should be zero (cf. Bryson and Ho)
l_end = X(3:4);

% Formulate root-finding problem
prob = struct('x', l_init, 'f', 0, 'g', l_end);
solver = nlpsol('solver', 'ipopt', prob);

% Solve the problem
arg = struct('x0', 0, 'lbg', 0, 'ubg', 0);
sol = solver(arg);
l_init = full(sol.x);

% Simulate the system
M = 100;
DT = T/M;
aug_opt = [x_init; l_init];
for j=1:M
  Fj = F(struct('x0', aug_opt(:, end), 'p', DT));
  aug_opt = [aug_opt, full(Fj.xf)];
end

% Calculate the optimal control
u_fcn = Function('u_fcn', {aug}, {u_opt});
u_opt = [];
for j=1:M+1
  r = u_fcn({aug_opt(:,j)});
  u_opt = [u_opt, full(r{1})];
end

% Plot the results
tgrid = linspace(0, T, M+1);
figure(1)
clf
hold on
plot(tgrid, aug_opt(1, :), '--')
plot(tgrid, aug_opt(2, :), '-')
plot(tgrid, u_opt, '-.')
title('Indirect single shooting')
xlabel('time')
legend('x1','x2','u')
grid on
