from casadi import *

# Time horizon
T = 10.

# State
x1=SX.sym('x1')
x2=SX.sym('x2')
x = vertcat((x1,x2))

# Control
u = SX.sym('u')

# ODE right hand side
xdot = vertcat([(1 - x2**2)*x1 - x2 + u, x1])

# Quadratic objective
L = x1**2 + x2**2 + u**2

# Costate
lam = SX.sym('lam',2)

# Hamiltonian function
H = dot(lam,xdot) + L
print 'H = ', H

# Costate equations
ldot = -gradient(H,x)

# H is of a convex quadratic form in u: H = u*u + p*u + q, let's get the coefficient p
p = gradient(H,u)     # this gives us 2*u + p
p = substitute(p,u,0) # replace u with zero: gives us p

# H's unconstrained minimizer is: u = -p/2
u_opt = -p/2

# We must constrain u to the interval [-1.0, 1.0], convexity of H ensures that the optimum is obtain at the bound when u_opt is outside the interval
u_opt = fmin(u_opt,  1.0)
u_opt = fmax(u_opt, -1.0)
print 'u_opt = ', u_opt

# Augment the state and substitute in the value for the optimal control
aug = vertcat((x,lam))
augdot = vertcat((xdot,ldot))
augdot = substitute(augdot,u,u_opt)

# Create an integrator (CVODES)
tf = SX.sym('tf')
dae = {'x':aug, 'p':tf, 'ode':tf*augdot}
opts = {'abstol':1e-8, 'reltol':1e-8}
F = integrator('F', 'cvodes', dae, opts)

# The initial state
x_init = [0.,1.]

# The initial costate
l_init = MX.sym('l_init',2)

# Get an expression for the augmented state at the end time
Fj = F({'x0':vertcat((x_init, l_init)), 'p':T});
X = Fj['xf']

# Costate at the final time should be zero (cf. Bryson and Ho)
l_end = X[2:4]

# Formulate root-finding problem
prob = {'x':l_init, 'f':0, 'g':l_end}
solver = nlpsol('solver', 'ipopt', prob)

# Solve the problem
arg = {'x0':0, 'lbg':0, 'ubg': 0}
sol = solver(arg)
l_init = sol['x']

# Simulate the system
M = 100
DT = T/M
aug_opt = [vertcat((x_init, l_init))]
for j in range(M):
   Fj = F({'x0': aug_opt[-1], 'p':DT})
   aug_opt += [ Fj['xf'] ]

# Calculate the optimal control
u_fcn = Function('u_fcn', [aug], [u_opt]);
u_opt = []
for v in aug_opt:
   [u] = u_fcn([v])
   u_opt += [u]

# Collect the optimal solution
aug_opt = horzcat(aug_opt).T
u_opt = horzcat(u_opt).T
tgrid = linspace(0, T, M+1)

# Plot the results
import matplotlib.pyplot as plt
plt.figure(1)
plt.clf()
plt.plot(tgrid, aug_opt[:, 0], '--')
plt.plot(tgrid, aug_opt[:, 1], '-')
plt.plot(tgrid, u_opt, '-.')
plt.title('Indirect single shooting')
plt.xlabel('time')
plt.legend(['x1','x2','u'])
plt.grid()
plt.show()
